﻿#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2012 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using System;
using System.Collections.Generic;
using System.Linq;
using HeuristicLab.Algorithms.GeneticAlgorithm;
using HeuristicLab.Common;
using HeuristicLab.Core;
using HeuristicLab.Data;
using HeuristicLab.Optimization;
using HeuristicLab.Parameters;
using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;
using HeuristicLab.Problems.DataAnalysis;
using HeuristicLab.Problems.DataAnalysis.Symbolic;
using HeuristicLab.Random;

namespace HeuristicLab.Algorithms.DataAnalysis.Symbolic {
  [Item("Symbolic DataAnalysis Island Genetic Algorithm", "A symbolic data analysis island genetic algorithm.")]
  [Creatable("Algorithms")]
  [StorableClass]
  public sealed class SymbolicDataAnalysisIslandGeneticAlgorithm : IslandGeneticAlgorithm {
    private const string FixedSamplesParameterName = "NumberOfFixedSamples";
    private const string FixedSamplesPartitionParameterName = "FixedSamplesPartition";
    private const string FixedSamplesPartitionsParameterName = "FixedSamplesPartitions";
    private const string RandomSamplesParameterName = "NumberOfRandomSamples";
    private const string EvaluatorParameterName = "IslandEvaluator";
    private const string ProblemEvaluatorParameterName = "ProblemEvaluator";

    #region Problem Properties
    public override Type ProblemType {
      get { return typeof(ISymbolicDataAnalysisSingleObjectiveProblem); }
    }
    public new ISymbolicDataAnalysisSingleObjectiveProblem Problem {
      get { return (ISymbolicDataAnalysisSingleObjectiveProblem)base.Problem; }
      set { base.Problem = value; }
    }
    #endregion

    #region parameters
    public IFixedValueParameter<IntValue> FixedSamplesParameter {
      get { return (IFixedValueParameter<IntValue>)Parameters[FixedSamplesParameterName]; }
    }
    public IValueParameter<ItemArray<IntRange>> FixedSamplesPartitionsParameter {
      get { return (IValueParameter<ItemArray<IntRange>>)Parameters[FixedSamplesPartitionsParameterName]; }
    }
    public IFixedValueParameter<IntValue> RandomSamplesParameter {
      get { return (IFixedValueParameter<IntValue>)Parameters[RandomSamplesParameterName]; }
    }
    public IValueParameter<ISymbolicDataAnalysisIslandGAEvaluator> EvaluatorParameter {
      get { return (IValueParameter<ISymbolicDataAnalysisIslandGAEvaluator>)Parameters[EvaluatorParameterName]; }
    }
    private ILookupParameter<ISingleObjectiveEvaluator> ProblemEvaluatorParameter {
      get { return (ILookupParameter<ISingleObjectiveEvaluator>)Parameters[ProblemEvaluatorParameterName]; }
    }
    #endregion

    #region properties
    public int FixedSamples {
      get { return FixedSamplesParameter.Value.Value; }
      set { FixedSamplesParameter.Value.Value = value; }
    }
    public ItemArray<IntRange> FixedSamplesPartitions {
      get { return FixedSamplesPartitionsParameter.Value; }
      set { FixedSamplesPartitionsParameter.Value = value; }
    }
    public int RandomSamples {
      get { return RandomSamplesParameter.Value.Value; }
      set { RandomSamplesParameter.Value.Value = value; }
    }
    #endregion

    [StorableConstructor]
    private SymbolicDataAnalysisIslandGeneticAlgorithm(bool deserializing) : base(deserializing) { }
    [StorableHook(HookType.AfterDeserialization)]
    private void AfterDeserialization() {
      RegisterParameterEvents();
    }
    private SymbolicDataAnalysisIslandGeneticAlgorithm(SymbolicDataAnalysisIslandGeneticAlgorithm original, Cloner cloner)
      : base(original, cloner) {
      RegisterParameterEvents();
    }
    public override IDeepCloneable Clone(Cloner cloner) {
      return new SymbolicDataAnalysisIslandGeneticAlgorithm(this, cloner);
    }

    public SymbolicDataAnalysisIslandGeneticAlgorithm()
      : base() {
      Parameters.Add(new FixedValueParameter<IntValue>(FixedSamplesParameterName, "The number of fixed samples used for fitness calculation in each island.", new IntValue(0)));
      Parameters.Add(new ValueParameter<ItemArray<IntRange>>(FixedSamplesPartitionsParameterName, "The fixed samples partitions used for fitness calculation for every island."));
      Parameters.Add(new FixedValueParameter<IntValue>(RandomSamplesParameterName, "The number of random samples used for fitness calculation in each island.", new IntValue(0)));
      Parameters.Add(new OptionalValueParameter<ISymbolicDataAnalysisIslandGAEvaluator>(EvaluatorParameterName, "The evaluator of the algorithm."));
      Parameters.Add(new LookupParameter<ISingleObjectiveEvaluator>(ProblemEvaluatorParameterName, "Internal parameter for name translation", "Evaluator"));

      ScopeTreeAssigner<IntRange> fixedSamplesPartitionCreator = new ScopeTreeAssigner<IntRange>();
      fixedSamplesPartitionCreator.LeftSideParameter.ActualName = FixedSamplesPartitionParameterName;
      fixedSamplesPartitionCreator.RightSideParameter.ActualName = FixedSamplesPartitionsParameterName;

      RandomCreator insertionPoint = OperatorGraph.Iterate().OfType<RandomCreator>().Skip(1).First();
      fixedSamplesPartitionCreator.Successor = insertionPoint.Successor;
      insertionPoint.Successor = fixedSamplesPartitionCreator;

      RegisterParameterEvents();
      RecalculateFixedSamplesPartitions();
    }

    protected override void Problem_EvaluatorChanged(object sender, EventArgs e) {
      ParameterizeProblemEvaluator();
      base.Problem_EvaluatorChanged(sender, e);
    }

    private void ParameterizeProblemEvaluator() {
      var regresssionEvaluator = Problem.Evaluator as ISymbolicDataAnalysisEvaluator<IRegressionProblemData>;
      if (regresssionEvaluator != null) {
        regresssionEvaluator.EvaluationPartitionParameter.ActualName = FixedSamplesPartitionParameterName;
      }
      var classificationEvaluator = Problem.Evaluator as ISymbolicDataAnalysisEvaluator<IClassificationProblemData>;
      if (classificationEvaluator != null) {
        classificationEvaluator.EvaluationPartitionParameter.ActualName = FixedSamplesPartitionParameterName;
      }
    }

    protected override void ParameterizeSolutionsCreator() {
      base.ParameterizeSolutionsCreator();
      SolutionsCreator.EvaluatorParameter.ActualName = EvaluatorParameterName;
    }

    protected override void ParameterizeMainLoop() {
      base.ParameterizeMainLoop();
      MainLoop.EvaluatorParameter.ActualName = EvaluatorParameterName;
      MainLoop.QualityParameter.ActualName = EvaluatorParameter.Value.QualityParameter.ActualName;
    }

    private void RegisterParameterEvents() {
      NumberOfIslandsParameter.ValueChanged += NumberOfIslandsParameter_ValueChanged;
      NumberOfIslandsParameter.Value.ValueChanged += (o, ev) => RecalculateFixedSamplesPartitions();
      FixedSamplesParameter.Value.ValueChanged += (o, e) => RecalculateFixedSamplesPartitions();
    }

    private void NumberOfIslandsParameter_ValueChanged(object sender, EventArgs e) {
      NumberOfIslands.ValueChanged += (o, ev) => RecalculateFixedSamplesPartitions();
      RecalculateFixedSamplesPartitions();
    }

    protected override void Problem_Reset(object sender, EventArgs e) {
      RecalculateFixedSamplesPartitions();
      ParameterizeProblemEvaluator();
      base.Problem_Reset(sender, e);
    }

    protected override void OnProblemChanged() {
      Problem.FitnessCalculationPartition.ValueChanged += (o, e) => RecalculateFixedSamplesPartitions();

      ParameterizeProblemEvaluator();
      if (Problem is IRegressionProblem) {
        var evaluator = new SymbolicDataAnalysisIslandGAEvaluator<IRegressionProblemData>();
        evaluator.FixedSamplesParameter.ActualName = FixedSamplesParameterName;
        evaluator.RandomSamplesParameter.ActualName = RandomSamplesParameterName;
        EvaluatorParameter.Value = evaluator;
      } else if (Problem is IClassificationProblem) {
        var evaluator = new SymbolicDataAnalysisIslandGAEvaluator<IClassificationProblemData>();
        evaluator.FixedSamplesParameter.ActualName = FixedSamplesParameterName;
        evaluator.RandomSamplesParameter.ActualName = RandomSamplesParameterName;
        EvaluatorParameter.Value = evaluator;
      } else
        EvaluatorParameter.Value = null;

      ParameterizeStochasticOperatorForIsland(EvaluatorParameter.Value);

      RecalculateFixedSamplesPartitions();
      base.OnProblemChanged();
    }

    private void RecalculateFixedSamplesPartitions() {
      if (Problem == null) {
        FixedSamplesPartitions = new ItemArray<IntRange>(Enumerable.Repeat(new IntRange(), NumberOfIslands.Value));
        return;
      }
      var samplesStart = Problem.FitnessCalculationPartition.Start;
      var samplesEnd = Problem.FitnessCalculationPartition.End;
      var totalSamples = Problem.FitnessCalculationPartition.Size;
      var fixedSamples = FixedSamples;
      var islands = NumberOfIslands.Value;

      int offset = 0;
      //fixed samples partition do not overlap
      if (((double)totalSamples) / fixedSamples <= islands) {
        offset = totalSamples / islands;
      } else {
        offset = (totalSamples - fixedSamples) / (islands - 1);
      }

      List<IntRange> partitions = new List<IntRange>();
      for (int i = 0; i < islands; i++) {
        var partitionStart = samplesStart + offset * i;
        partitions.Add(new IntRange(partitionStart, partitionStart + fixedSamples));
      }

      //it can be the case that the last partitions exceeds the allowed samples
      //move the last partition forward.
      int exceedsSamples = partitions[partitions.Count - 1].End - samplesEnd;
      if (exceedsSamples > 0) {
        partitions[partitions.Count - 1].Start -= exceedsSamples;
        partitions[partitions.Count - 1].End -= exceedsSamples;
      }
      FixedSamplesPartitions = new ItemArray<IntRange>(partitions);
    }

  }
}
